import oscP5.*;
import peasy.*;
import processing.core.*;


/*
 * Madeline Gannon _ www.madlab.cc
 * 
 * 01.28.2012   |   Grasshopper2Processing Demo   |   @madelinegannon
 * 
 * Framework for bringing 3D mesh data from Rhino (via Grasshopper) into
 * Processing_0195.  The mesh from Rhino is exported as .txt files that
 * contain the model's UNIQUE VERTICES, as well as the ORDERED VERTICES 
 * needed to construct each face.  .txt are streamed from the Rhino/GH 
 * file, into the sketch's data folder.
 * 
 * This demo uses FaceOSC to control a 3D puppet, using eyebrow and mouth 
 * data to modify the ears and jaw. The code synchronizes the movements of 
 * your head, mouth, and eyebrows to control the expressions of a tessellated 
 * three-dimensional model. The model was first developed in Rhino3D, then 
 * translated to the Processing programming environment through Grasshopper, 
 * then connected to Kyle McDonalds FaceOSC app through Oscp5.  PeasyCam is 
 * also used for navigation.
 * 
 * Pivot points, as well as points to be rotated are specified in Grasshopper, 
 * then exported as .txt files to Processing.  While the current puppet is 
 * highly tessellated, one can begin to create additional faces 
 * (and unique vertices) to give the model higher detail.
 * 
 * FaceOSC can be downloaded here: 
 * 		https://github.com/kylemcdonald/ofxFaceTracker/downloads
 * 
 * The MyFace class is a template from Dan Wilcox: 
 * 		https://github.com/danomatika/Face-OSC-Templates
 * 
 */


public class DigitalPuppet extends PApplet{

	OscP5 oscP5;
	PeasyCam cam;

	// our FaceOSC tracked face data
	MyFace face = new MyFace();
	
	// UNIQUE VERTICES from GH
	PVector[] vertices;	

	// holds the indices into the UNIQUE VERTICES, vertices
	int[] vertIndex;
	int[] earIndex;
	int[] jawIndex;
	

	// to track whether eye/jaw is opening or closing 
	float prevMouthWidth = 0;
	float prevMouthHeight = 0;
	float prevEyebrow = 0;

	// to smooth incoming position/orientation values
	float prevPoseScale    = 0;
	float prevXorientation = 0;
	float prevYorientation = 0;
	float prevZorientation = 0;
	

	public void setup() {
	  background(255);
	  size(900, 600, P3D);
	  smooth();

	  oscP5 = new OscP5(this, 8338);

	  cam = new PeasyCam(this, 100);
	  cam.setMinimumDistance(5);
	  cam.setMaximumDistance(1500);

	  
	  // Load mesh data from GH and convert to PVector
	  String[] ordering = loadStrings("faceOSC_model_vertexOrdering.txt");
	  PVector[] pts = new PVector[ordering.length];	  
	  processGH(ordering,pts);
	  
	  String[] uniqueVertices = loadStrings("faceOSC_model_uniqueVertices.txt");
	  vertices = new PVector[uniqueVertices.length];	  
	  processGH(uniqueVertices,vertices);
	  	  
	  String[] ear = loadStrings("faceOSC_model_earVertices.txt");
	  PVector[] ears = new PVector[ear.length];	
	  processGH(ear,ears);
	  
	  String[] jaw = loadStrings("faceOSC_model_jawVertices.txt");
	  PVector[] jaws = new PVector[jaw.length];	
	  processGH(jaw,jaws);
	  
	  
	  // make the ORDER FACE VERTICES reference UNIQUE VERTICES
	  vertIndex = new int[ordering.length];	  
	  vertIndex = resequencePoints(pts, vertices);
	  
	  // make EAR VERTICES reference UNIQUE VERTICES
	  earIndex = new int[ear.length];
	  earIndex = resequencePoints(ears,vertices);
	  
	  // make JAW VERTICES reference UNIQUE VERTICES
	  jawIndex = new int[jaw.length];
	  jawIndex = resequencePoints(jaws,vertices);
	  	  
	}


	public void draw() {
	  background(0);
	  lights();
	  
	  pushMatrix();
	  
	  scale(prevPoseScale);
	  rotateX(1.4f-prevXorientation);
	  rotateY(-1*prevYorientation);
	  rotateZ(3.2f+prevZorientation);
	  
	  // average previous and current pose/orientation to smooth values
	  prevPoseScale = (prevPoseScale + face.poseScale)/2;
	  prevXorientation = (prevXorientation + face.poseOrientation.x)/2;
	  prevYorientation = (prevYorientation + face.poseOrientation.z)/2;
	  prevZorientation = (prevZorientation + face.poseOrientation.y)/2;


	  /*
	  // temp axes
	  strokeWeight(1.5f);	  
	  stroke(0, 189, 255);			// z = blue
	  line(0, 0, 0, 0, 0, 20);	  
	  stroke(152, 255, 0);			// y = green
	  line(0, 0, 0, 0, 10, 0);	  
	  stroke(255, 196, 0);			// x = orange
	  line(0, 0, 0, 10, 0, 0);
	  */
	  
	  drawFace();
	  
	  // mirror the face and redraw
	  pushMatrix();
	  scale(-1, 1, 1);
	  drawFace();
	  popMatrix();
	  
	  popMatrix();

	  if (face.found > 0) { 		  
		  moveJaw();
		  moveEars();
	  }
	  
	  
	}


	/*
	 *  OSC CALLBACK FUNCTIONS
	 */
	public void oscEvent(OscMessage m) {
	  face.parseOSC(m);
	}
	
	
	/*
	 * Rewrite the redundant values of 'points' to reference unique values
	 * from the 'reference' array.  
	 * 
	 * Store the index values for reconstructing mesh from the reference vertices.
	 */
	public int[] resequencePoints(PVector[] points, PVector[] ref){
		
		int[] index = new int[points.length];
		
		// for every point in the ordered array, find it's unique vertex with 
		// the vertices array, and store it's index
		int c=0;
		for (int i=0; i<points.length; i++){
			
			for (int j=0; j<ref.length; j++){
				 
				if (points[i].x == ref[j].x && 
					points[i].y == ref[j].y &&
					points[i].z == ref[j].z){
					index[i] = j;	
					c++;
				}
			}					
		}

		return index;
	}
	
	
	/*
	 * Takes a .txt file from grasshopper and transforms
	 * it to an array of PVectors	 
	 */
	public void processGH(String[] coords, PVector[]points){
		
		for (int i=0; i<points.length; i++) {
		    // parse string for XYZ coords, and convert to float for PVector
		    String[] temp = split(coords[i], ", ");
		    float x = Float.parseFloat(temp[0]);
		    float y = Float.parseFloat(temp[1]);
		    float z = Float.parseFloat(temp[2]);
		    
		    // round each point to 3 decimal places
		    x = round(x,3);
		    y = round(y,3);
		    z= round(z,3);
		    
		    points[i] = new PVector(x, y, z);
		}
		
	}
	
	/*
	 * Round the incoming points to a fixed decimal 
	 */
	private float round(float num, int decimal) {
		  float scalar = (float)Math.pow(10,decimal);
		  num = num * scalar;
		  float temp = Math.round(num);
		  return (float)temp/scalar;
		  }
	
	
	
	/*
	 * Use vertIndex to index into vertices[] and reconstruct the mesh. 
	 * ... every 3 points make one mesh face.
	 * 
	 */
	public void drawFace() {

		fill(200, 200, 200, 220);
		stroke(250, 250, 250, 200);

		
		for (int i=0; i<vertIndex.length; i+=3){
			beginShape();
			for (int j=0; j<3; j++)
				vertex(vertices[vertIndex[i+j]].x,vertices[vertIndex[i+j]].y,vertices[vertIndex[i+j]].z);
			
			endShape(CLOSE);
			
		}		

	}
	
	/*
	 * Rotate the ears about a given pivot point along the Y-axis. 
	 * 
	 * The first point in the list is the pivot point ... DO NOT ROTATE.
	 * 
	 * Map the angle of rotation from the change in eyebrow movement. 
	 */
	public void moveEars() {
		
		PVector pivot = vertices[earIndex[0]];

		float rot = 0; 
		float currEyebrow = map(face.eyebrowLeft, 7, 10, -1, 1);
		
		// smooth incoming eyebrow values
		prevEyebrow = (prevEyebrow + currEyebrow)/2;
		
		// raising eyebrow
		if (face.eyebrowLeft > prevEyebrow) {
			rot += prevEyebrow - currEyebrow;
			prevEyebrow = currEyebrow;
		}

		// lowering eyebrow
		else if (face.eyebrowLeft < prevEyebrow) {
			rot -= currEyebrow - prevEyebrow;
			prevEyebrow = currEyebrow;
		}


		float scalar = -.2f;

		// PIVOT POINT AT i==0
		for (int i=1; i<earIndex.length; i++) {

			// hard-coded tweaking for effect
			if (i == 1)
				scalar = -.1f;
			else if (i == 3)
				scalar = -.16f;
			else
				scalar = -.3f;

			// translate the point the rotational axis
			PVector temp = vertices[earIndex[i]];
			temp.x += pivot.x;
			temp.z += pivot.z;

			// perform rotation about Y axis
			float z = temp.z*cos(scalar*rot) - temp.x*sin(scalar*rot);    
			float x = temp.z*sin(scalar*rot) + temp.x*cos(scalar*rot);

			// un-translate points
			x -= pivot.x;
			z -= pivot.z;    

			// modify the original vertex by the new X and Z location
			vertices[earIndex[i]] = new PVector(x, temp.y, z);

		}
	}

	
	/*
	 * Rotate the jaw about a given pivot point along the X-axis. 
	 * 
	 * The first point in the list is the pivot point ... DO NOT ROTATE.
	 * 
	 * Map the angle of rotation from the change in mouth movement. 
	 */
	public void moveJaw() {

		PVector pivot = vertices[jawIndex[0]];
		
		float currMouthWidth = map(face.mouthWidth,6,16,0,1);
		float dy = 0;
		
		// smooth incoming mouthWidth values
		prevMouthWidth = (prevMouthWidth + currMouthWidth)/2;
				
		
		// mouth is widening
		if (face.mouthWidth > prevMouthWidth) {
			dy += prevMouthWidth - currMouthWidth;
			prevMouthWidth = currMouthWidth;
		}

		// mouth is pursing
		else if (face.mouthWidth < prevMouthWidth) {
			dy -= currMouthWidth - prevMouthWidth;
			prevMouthWidth = currMouthWidth;
		}

		// modify mouth's pivot point to create 'smile'
		pivot.y += (-1.5*dy);
		pivot.x += (dy);


		// open and close the mouth
		float rot = 0;  
		float currMouthHeight = map(face.mouthHeight, 0, 9, 0, 1);
		
		// smooth incoming mouthHeight values
		prevMouthHeight = (prevMouthHeight + currMouthHeight)/2;
		
		// opening mouth
		if (face.mouthHeight > prevMouthHeight) {
			rot += prevMouthHeight - currMouthHeight;
			prevMouthHeight = currMouthHeight;
		}

		//closing mouth
		else if (face.mouthHeight < prevMouthHeight) {
			rot -= currMouthHeight - prevMouthHeight;
			prevMouthHeight = currMouthHeight;
		}

		float scalar;

		// PIVOT POINT AT i==0
		for (int i=1; i<jawIndex.length; i++) {
			
			// hard-coded tweaking for effect
			if (i == 5 || i == 5)
				scalar = -.3f;
			else
				scalar = -.6f;

			// translate the point the rotational axis
			PVector temp = vertices[jawIndex[i]];
			temp.y += pivot.y;
			temp.z += pivot.z;

			// perform rotation about X axis
			float y = temp.y*cos(scalar*rot) - temp.z*sin(scalar*rot);    
			float z = temp.y*sin(scalar*rot) + temp.z*cos(scalar*rot);

			// un-translate points
			y -= pivot.y;
			z -= pivot.z;

			// modify the original vertex by the new Y and Z location
			vertices[jawIndex[i]] = new PVector(temp.x, y, z);
		}
	}
	

}
